@@ -24,7 +24,6 @@
 #include <unistd.h>
 #include <getopt.h>
 #include <string.h>
-#include <strings.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <sys/types.h>
@@ -38,6 +37,7 @@
 #include "nls.h"
 #include "widechar.h"
 #include "strutils.h"
+#include "closestream.h"
 
 #ifndef MAXSYMLINKS
 #define MAXSYMLINKS 256
@@ -123,7 +123,7 @@ add_id(struct idcache **ic, char *name, unsigned long int id, int *width)
 	/* note, we ignore names with non-printable widechars */
 	if (w > 0)
 		nc->name = xstrdup(name);
-	else if (asprintf(&nc->name, "%lu", id) == -1)
+	else if (xasprintf(&nc->name, "%lu", id) == -1)
 		nc->name = NULL;
 
 	for (x = *ic; x && x->next; x = x->next);
@@ -179,6 +179,7 @@ readlink_to_namei(struct namei *nm, const char *path)
 {
 	char sym[PATH_MAX];
 	ssize_t sz;
+	int isrel = 0;
 
 	sz = readlink(path, sym, sizeof(sym));
 	if (sz < 1)
@@ -186,20 +187,26 @@ readlink_to_namei(struct namei *nm, const char *path)
 	if (*sym != '/') {
 		char *p = strrchr(path, '/');
 
-		nm->relstart = p ? p - path : 0;
-		if (nm->relstart)
+		if (p) {
+			isrel = 1;
+			nm->relstart = p ? p - path : 0;
 			sz += nm->relstart + 1;
+		}
 	}
 	nm->abslink = xmalloc(sz + 1);
 
-	if (*sym != '/' && nm->relstart) {
+	if (*sym != '/' && isrel) {
 		/* create the absolute path from the relative symlink */
 		memcpy(nm->abslink, path, nm->relstart);
 		*(nm->abslink + nm->relstart) = '/';
 		nm->relstart++;
 		memcpy(nm->abslink + nm->relstart, sym, sz - nm->relstart);
 	} else
+		/* - absolute link (foo -> /path/bar)
+		 * - or link without any subdir (foo -> bar)
+		 */
 		memcpy(nm->abslink, sym, sz);
+
 	nm->abslink[sz] = '\0';
 }
 
@@ -221,7 +228,7 @@ dotdot_stat(const char *dirname, struct stat *st)
 	memcpy(path + len, DOTDOTDIR, sizeof(DOTDOTDIR));
 
 	if (stat(path, st))
-		err(EXIT_FAILURE, _("could not stat '%s'"), path);
+		err(EXIT_FAILURE, _("stat failed %s"), path);
 	free(path);
 	return st;
 }
@@ -448,6 +455,7 @@ main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
+	atexit(close_stdout);
 
 	while ((c = getopt_long(argc, argv, "hVlmnovx", longopts, NULL)) != -1) {
 		switch(c) {
